/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DRIVERS_HPP_
#define DRIVERS_HPP_

#include "compiler.hpp"
#include "keyfile.hpp"
#include "sessions.hpp"
#include "logger.hpp"
#include "utils.hpp"

namespace server {
class Driver {
friend class Session;
friend class script;
public:
    using context_t = script::context_t;
    using header_t = script::header_t;
    using check_t = const char *(*)(const script::context_t& scr);
    using line_t = script::line_t;

    struct keyword_t {
        method_t method;
        check_t check;

        inline keyword_t() {
            method = nullptr;
            check = nullptr;
        }

        inline keyword_t(method_t m, check_t c) {
            method = m;
            check = c;
        }

        inline operator bool() const {
            return method != nullptr;
        }

        inline auto operator!() const {
            return method == nullptr;
        }
    };

    Driver(const Driver&) = delete;
    auto operator=(const Driver&) -> Driver& = delete;

    virtual ~Driver();

    inline auto name() const {
        return name_;
    }

    inline auto keys() const {
        return keys_;
    }

    inline static auto instance() {
        return driver_;
    }

    inline static auto get_script(const std::string& name) {
        std::lock_guard<std::mutex> lock(script_lock);
        return scripts[name];
    }

    inline static auto count() {
        std::lock_guard<std::mutex> lock(session_lock);
        return sessions.size();
    }

    inline static auto get(std::string_view sid) {
        std::lock_guard<std::mutex> lock(session_lock);
        return sessions.at(sid);
    }

    inline static auto extend_runtime(const std::string& id, const keyword_t& keyword) {
        keywords[id] = keyword;
    }

    inline static auto extend_runtime(const std::initializer_list<std::pair<std::string, keyword_t>>& list) {
        for(auto const& [key, value] : list)
            keywords[key] = value;
    }

    NO_EXPORT static auto config_instance(keyfile cfg) -> keyfile;
    NO_EXPORT static void reload_instance(const keyfile& cfg);
    NO_EXPORT static void start_instance(const keyfile& ini);
    NO_EXPORT static void drop_instance();
    NO_EXPORT static void stop_instance();

    static void attach(session_t& s);
    static void release(std::string_view sid);
    static auto control(const std::vector<std::string_view>& list) -> bool;

protected:
    static std::unordered_map<std::string_view, keyword_t> keywords;
    static std::unordered_map<std::string_view, session_t> sessions;
    static std::unordered_map<std::string_view, script> scripts;
    static std::unordered_map<std::string_view, const char *> groups;
    static std::mutex session_lock;
    static std::mutex script_lock;

    explicit Driver(const std::string& id) noexcept;

    virtual void startup(const keyfile& config) = 0;
    virtual void restart(const keyfile& config);
    virtual void configure(keyfile config);
    virtual void shutdown();
    virtual void disconnect();

    // check and coding related functions

    inline static auto keyword(const std::string& key) {
        return keywords[key];
    }

    inline static auto check(check_t handler, const context_t& ctx) {
        return (*(handler))(ctx);
    }

    inline static auto event_group(const char *id) {
        auto result = groups[id];
        return result ? result : id;
    }

private:
    std::string name_;
    keyfile::keys keys_;

    static Driver *driver_;
};

extern volatile bool running;		// global server run flag
extern bool ipv6;					// global ipv6 setting
} // end namespace

namespace this_thread {
using namespace std::this_thread;

auto get_priority() -> int;
auto set_priority(int priority) -> bool;
} // end namespaces

#ifdef BUILD_STATIC
// a driver we can force load for static builds...
namespace loader {
extern server::Driver *bind_driver();
}
#endif

/*!
 * Base class for deriving and managing plugin driver.
 * \file drivers.hpp
 */
#endif
