/*
 * Copyright (C) 2017-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef RTP_HPP_
#define RTP_HPP_

#include "memory.hpp"
#include "clock.hpp"

#include <thread>
#include <mutex>
#include <stdexcept>
#include <string_view>
#include <unordered_map>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>

namespace server {
static const size_t rtp_min_header_size = 12;
static const size_t rtp_max_header_size = 12 + (15 * 4); // 15 csrc...
static const size_t max_audio_size = 160;
class Session;

struct audio_data_t final {
    uint8_t *data;
    size_t size;
};

class rtp_error final : public std::exception
{
public:
    explicit rtp_error(const char *text, int code = 0) noexcept {
        err = code;
        msg = text;
    }

    auto what() const noexcept -> const char * override {
        return msg;
    }

    inline auto error() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

class rtp_packet final {
public:
    explicit rtp_packet(uint8_t *p) :
    packet(p), samples_(0), framing_(20) {}

    inline auto version() const {  // default is 80/2
        return (packet[0] & 0xc0) >> 6;
    }

    inline auto padding() const {
        return (packet[0] & 0x20) != 0;
    }

    inline auto contrib() const {
        return (packet[0] & 0x0f);
    }

    inline auto marker() const {
        return (packet[1] & 0x80) != 0;
    }

    inline auto payload() const {
        return (packet[1] & 0x7f);
    }

    inline auto sequence() const {
        return get16(&packet[2]);
    }

    inline auto timestamp() const {
        return get32(&packet[4]);
    }

    inline auto source() const {
        return get32(&packet[8]);
    }

    inline auto extends() const -> uint16_t {
        if(!(packet[0] & 0x10))
            return 0;
        return get16(&packet[rtp_min_header_size + (contrib() * 4) + 2]) + 4;
    }

    inline auto profile_id() const -> uint16_t {
        if(!extends())
            return 0;
        return get16(&packet[rtp_min_header_size + (contrib() * 4)]);
    }

    inline auto profile_size() const -> uint16_t {
        if(!extends())
            return 0;
        return get16(&packet[rtp_min_header_size + (contrib() * 4) + 2]);
    }

    inline auto header_size() const {
        return  rtp_min_header_size + (contrib() * 4) + extends();
    }

    inline auto framing() const {
        return framing_;
    }

    inline auto samples() const {
        return samples_;
    }

    inline auto data() {
        return &packet[rtp_min_header_size + (contrib() * 4) + extends()];
    }

    inline auto profile_data() {
        return (extends() ? &packet[rtp_min_header_size + (contrib() * 4) + 4] : nullptr);
    }

    inline void set_payload(uint8_t value, uint32_t spf, milliseconds_t ft) {
        value |= (packet[1] & 0x80);
        packet[1] = value;
        samples_ = spf;
        framing_ = ft;
    }

    inline void set_version(uint8_t value = 2) {
        packet[0] = (value << 6) | (packet[0] & 0x3f);
    }

    inline void inc_sequence(unsigned count = 1) {
        set16(&packet[2], sequence() + count);
        set32(&packet[4], timestamp() + samples_ * count);
    }

    inline void inc_silence(unsigned count = 1) {
        set32(&packet[4], timestamp() + samples_ * count);
    }

    inline auto send(int socket, const struct sockaddr *addr, size_t size) {
        return ::sendto(socket, packet, size + header_size(), MSG_DONTWAIT, addr, addr_size(addr));
    }

    inline auto recv(int socket, struct sockaddr *addr, size_t max) {
        socklen_t len{sizeof(struct sockaddr_storage)};
        auto size = ::recvfrom(socket, packet, max, 0, addr, &len);
        if(size < 0)
            return size;
        if(size <= header_size())
            return ssize_t(0);
        return ssize_t(size - header_size());
    }

    inline static auto addr_size(const struct sockaddr *addr) -> socklen_t {
        switch(addr->sa_family) {
        case AF_INET:
            return sizeof(struct sockaddr_in);
        case AF_INET6:
            return sizeof(struct sockaddr_in6);
        default:
            return socklen_t(0);
        }
    }

private:
    inline static auto get16(const uint8_t *ptr) -> uint16_t {
        return (ptr[0] << 8) | ptr[1];
    }

    inline static auto get32(const uint8_t *ptr) -> uint32_t {
        return (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | ptr[3];
    }

    inline static void set16(uint8_t *ptr, uint16_t value) {
        ptr[0] = (value & 0xff00) >> 8;
        ptr[1] = value & 0xff;
    }

    inline static void set32(uint8_t *ptr, uint32_t value) {
        ptr[0] = (value & 0xff000000) >> 24;
        ptr[1] = (value & 0x00ff0000) >> 16;
        ptr[2] = (value & 0x0000ff00) >> 8;
        ptr[3] = value & 0xff;
    }

    uint8_t *packet;
    uint32_t samples_;
    milliseconds_t framing_;
};

class rtp_audio final {
public:
    std::unordered_map<std::string_view, uint8_t> payloads;

    explicit rtp_audio(Session *s);

    operator bool() const {
        return active;
    }

    auto operator!() const -> bool {
        return !active;
    }

    inline void join() {
        if(active) {
            thread.join();
            active = false;
        }
    }

    inline void sync_with_origin(bool sync_flag) {
        sync_address = sync_flag;
    }

    auto connect(const struct sockaddr *addr) -> bool;
    void disconnect();
    auto start();
    void finish();
    void silent();
    void ignore();
    void keepalive();

private:
    std::thread thread;         // exec thread
    std::mutex mutex;           // lock gaurd
    Session *session;           // session to post events to
    int stream_fd, record_fd, socket;
    struct sockaddr_storage remote_to, remote_from;
    bool sync_address;
    uint64_t sent_total, fail_total;
    uint32_t samples;           // samples per frame...
    milliseconds_t framing;
    uint8_t payload, dtmf;
    bool active;

    // memory loaded playback state
    audio_data_t *audio_data;
    unsigned audio_items, audio_offset;

    inline auto address() {
        return reinterpret_cast<struct sockaddr *>(&remote_to);
    }

    inline auto from() {
        if(!sync_address)
            return reinterpret_cast<struct sockaddr *>(&remote_from);
        return address();
    }

    auto send_output(rtp_packet& rtp) -> size_t;
    auto recv_input(rtp_packet& rtp, size_t max) -> size_t;
    void close_streams();
    void exec_thread();

    auto (rtp_audio::*put)(uint8_t *packet) noexcept -> size_t;
    auto put_silent(uint8_t *packet) noexcept -> size_t;
};
} // end namespace server

/*!
 * Basic rtp processing. This is used to access rtp header properties,
 * to perform basic media operations, etc.
 * \file rtp.hpp
 */
#endif
