/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SESSIONS_HPP_
#define SESSIONS_HPP_

#include "memory.hpp"
#include "symbols.hpp"
#include "clock.hpp"
#include "script.hpp"
#include "rtp.hpp"

#include <set>
#include <mutex>

namespace server {
enum event_type {
    timeout,        // timer expires...
    digit,          // dtmf digit..
    flash,          // flash event
    entry,          // entering new state...
    leave,          // leaving old state...
    exit,           // call exiting
    shutdown,       // system shutdown notification
};

class Session {
public:
    static const milliseconds_t step_rate = TUNE_STEP_RATE;
    static const unsigned step_count = TUNE_STEP_COUNT;
    static const size_t max_digits = TUNE_MAX_DIGITS;

    // this may be memory mapped in the future...
    struct stats_t {    // session statistics collected here...
        char sid[8];
        char state[8];
        system_time start;
        unsigned lines, steps;
    };

    struct event_t {
        event_type type;
        union {
            void *native_event;
            char digit[2];
            std::string_view sid;
        };
    };

    // Make sure if same shared pointer, always !<(l,r) && !(r,l)
    // If same time, we use memory order for uniqueness...
    // Otherwise ordered by time
    struct timer_compare {
        inline auto operator()(const session_t& lhs, const session_t& rhs) const -> bool {
            if(lhs.get() == rhs.get())
                return false;
            if(lhs->until() == rhs->until())
                return lhs.get() < rhs.get();
            return lhs->until() < rhs->until();
        }
    };

    using runlist_t = std::set<session_t, timer_compare>;
    using run_t = runlist_t::iterator;

    virtual ~Session();

    inline operator bool() const {
        return !sid_.empty();
    }

    inline auto operator!() const -> bool {
        return sid_.empty();
    }

    inline auto sid() const {
        return sid_;
    }

    inline auto until() const -> milliseconds_t {
        return expires.until();
    }

    inline auto get_symbol(const char *id) {
        std::lock_guard<std::mutex> lock(access_lock);
        return syms.get_value(id);
    }

    inline auto set_symbol(const char *id, const char *value) {
        std::lock_guard<std::mutex> lock(access_lock);
        return syms.set_value(id, value);
    }

    inline auto get_stats() const {
        return &stats;
    }

    static void release(const session_t& session);
    static void begin(const session_t& session);
    static auto send(const session_t& session, const event_t &event) -> bool;
    static void start(int priority);
    static void stop();

protected:
    friend class Driver;

    using handler_t = bool (Session::*)(const event_t&);

    stats_t stats;
    std::mutex access_lock;
    symbols syms;
    script image;
    handler_t handler;
    timer started, expires;
    char *digits;               // set from storage
    rtp_audio audio;            // rtp audio streaming...

    union {
        struct collect_t {
            size_t max_digits, min_digits;
            char *terminators;
            char *digits;
            milliseconds_t wait_until;
            milliseconds_t interdigit;
        } collect;
    } state;

    script::line_t *begins[script::max_levels];
    char temp[symbols::var_size + 1];
    const script::header_t *header;
    const script::trap_t *trap;
    const script::line_t *line;
    unsigned index, level;

    explicit Session(const script& attach) noexcept;

    inline auto process_event(const event_t& event) noexcept {
        return (this->*(handler))(event);
    }

    inline auto execute_method(method_t method) noexcept {
        return (this->*(method))();
    }

    inline auto argv() const {
        return line->argv;
    }

    inline void skip() {
        if(line)
            line = line->next;
        index = 0;
    }

    inline auto this_step() {
        return set_handler("script", &Session::script_handler);
    }

    inline auto next_step() {
        skip();
        return set_handler("script", &Session::script_handler);
    }

    inline void set_error(const char *text) {
        mem_copy(error_, symbols::var_size, text);
    }

    auto set_handler(const char *text, handler_t handler) -> bool;
    void set_timeout(const timer& until);
    void set_timeout(milliseconds_t offset);
    void bump_timeout(milliseconds_t offset);
    void clear_timeout();

    auto argv(unsigned short offset) -> const char *;
    auto format() -> const char *;   // test if currently a &xxx function
    auto option() -> const char *;   // test if currently in a =xx option
    auto symbol() -> const char *;   // get next symbol (var symname ...)
    auto value() -> const char *;    // get next argument value until nullptr

    virtual void begin();
    void entry();   // entry into new state, figure out event masks, etc...
    auto error(const char *reason) -> bool;
    auto event(const char *id) -> bool;
    auto label(const char *id) -> bool;

    virtual auto common_handler(const event_t &event) noexcept -> bool;
    auto collect_handler(const event_t &event) noexcept -> bool;
    auto sleep_handler(const event_t &event) noexcept -> bool;
    auto script_handler(const event_t &event) noexcept -> bool;
    auto dummy_handler(const event_t &event) noexcept -> bool;

private:
    run_t run_;
    char *error_;
    std::string_view sid_;

    auto funcs(const char *cp) -> const char *;       // ?xxx query ops...
};
} // end namespace

/*!
 * Base class for deriving Bayonne session instances.
 * \file sessions.hpp
 */
#endif
